module cgnorm_mod 


  implicit none
  real, dimension (:), allocatable, private   :: s, ss
  real, private      :: rt       ! rt = r_thr = r_bar for some norms 
  integer, private   :: ps_iter  ! No. of iteration on plane-search (alpha,beta)
  character(10)      :: ps_norm  ! Type of norm to be used for plane search


contains
  ! ========================================================================================
  !                   NORM Functions      ( RT needs to be set by calling CGNORM_INIT )
  ! ========================================================================================
  !!  Arguments IN: 
  !!     rr : residual (R)
  !!  Arguments OUT: 
  !!     c0 : Norm value at R                 (C )
  !!     c1 : Norm 1st derivative value at R  (C')
  !!     c2 : Norm 2nd derivative value at R  (C")
  
  !!  Output Flag : 
  !!     1 : Exit without completion
  !!     0 : Done without any exit or error   
  !!

  ! --------------------------------------------
  !       SIGNUM (L1) norm and Derivative
  ! --------------------------------------------
  integer function l1( rr, c0,c1,c2)
    !  Computes SIGNUM(L1) Norm and its 1st,2nd derivative around point rr
    real, dimension (:)            :: rr
    double precision, dimension(:) :: c0,c1,c2
    integer                        :: i
    
    l1=1
    do i=1,size(rr)
       
       c0(i)= abs(rr(i)) 
       c1(i)= rr(i)/max(abs(rr(i)),1.d-12)
       ! Implements the dirac delta function
       c2(i) = 2/(1e-6*sqrt(3.14159)) * exp( - (rr(i))**2 * 1e12 )  
    end do
      
    l1=0
  end function



  ! --------------------------------------------
  !       CLASSIC (L2) norm and Derivative
  ! --------------------------------------------
  integer function l2( rr, c0,c1,c2)
    !  Computes L2 (classic) Norm and its 1st,2nd derivative around point rr
    real, dimension (:)            :: rr
    double precision, dimension(:) :: c0,c1,c2
    integer                        :: i
    
    l2=1
    do i=1,size(rr)
       c0(i)= rr(i)*rr(i)/2
       c1(i)= rr(i)
       c2(i)= 1
    end do
    
    l2=0
  end function

  ! --------------------------------------------
  !         HYBRID norm and Derivative
  ! --------------------------------------------
  integer function hybrid( rr, c0,c1,c2)
    !  Computes HYBRID Norm and its 1st,2nd derivative around point rr
    ! C(0) = 0
    real, dimension (:)            :: rr
    double precision, dimension(:) :: c0,c1,c2
    double precision               :: term
    integer                        :: i
    
    hybrid=1
    do i=1,size(rr)
       !! Jon's HYBRID
       term = sqrt(1+rr(i)**2/(rt**2))
       c0(i)= rt**2 * (term -1) 
       c1(i)= rr(i)/term
       c2(i)= 1./(term**3)
    end do
    
    hybrid=0
  end function


  ! --------------------------------------------
  !         HUBER norm and Derivative
  ! --------------------------------------------
  integer function huber( rr, c0,c1,c2)
    !  Computes HUBER Norm and its 1st,2nd derivative around point rr
    !  Note that rt > 0 (set in init. function)
    real, dimension (:)            :: rr
    double precision, dimension(:) :: c0,c1,c2
    double precision               :: ratio
    integer                        :: i
    
    huber=1
    do i=1,size(rr)
       ratio = rr(i)/rt

       ! Jon's HUBER
       if (abs(ratio)>=1.) then
          c0(i)= abs(rr(i))-rt/2. 
          c1(i)= rr(i) / max(abs(rr(i)),1.d-12)
          c2(i) = 0.              ! Huber linear portion has 0 OR Inf. convexity 
          
       else
          c0(i)= rt* ratio**2/2.
          c1(i)= ratio
          c2(i)= 1/rt
       end if
       
    end do
    
    huber=0
  end function



  ! --------------------------------------------
  !    Derivative step for different norms
  ! --------------------------------------------
  integer function g_norm(rw,rd,norm,rthr)
    ! Compute g for different norms (rd,norm,rthr --> rw)
    !
    !!  Arguments IN: 
    !!     norm : nrom
    !!     rthr : RBAR/R_THR
    !!     rd   : -Wd +WFm0
    !!  Arguments OUT: 
    !!     rw   : C'(norm)*rd 
    
    !!  Output Flag : 
    !!     1 : Exit without completion
    !!     0 : Done without any exit or error   
    !!    
    character(10)       :: norm
    real                :: rthr
    real, dimension(:)  :: rd
    integer             :: i, stat

    real, dimension(:)  :: rw
    double precision, dimension(:), allocatable     :: c0,c1,c2
    allocate (c0(size(rd)),c1(size(rd)),c2(size(rd))) 
    
    g_norm=1
    ! Note that rt > 0 (set in init. function)
    rthr=abs(rthr)


    if      (norm(1:2)=="l2")    then
       ! L2 Norm
       rw  = rd

    else if (norm(1:2)=="l1")    then
       ! L1 Norm
       do i=1,size(rd)
          rw(i)  = rd(i)/max(abs(rd(i)),1.d-12)
       end do

    else if (norm(1:6)=="hybrid") then
       ! Hybrid Norm
       do i=1,size(rd)
          rw(i)  = rd(i)/( sqrt(1+rd(i)**2/rthr**2))
       end do


    else
       ! Huber Norm
       do i=1,size(rd)
          if (abs(rd(i))>=rthr) then
             rw(i) = rd(i)/abs(rd(i))
          else
             rw(i) = rd(i)/rthr
          endif
       end do

    endif



    g_norm=0
  end function








  ! ========================================================================================
  !                          CGNORM Plane Search function
  ! ========================================================================================
    ! --------------------------------------------
    !         Initialize CGNORM function
    ! --------------------------------------------
    subroutine cgnorm_init(rthr,norm,psiter)
      ! Initialize CGNORM Stepper with r_thr, norm type and psiter_in
      !
      ! Arguments IN: 
      !    rthr      : RBAR or R_THR for norm 
      !    norm_in   : Norm ype to be used (Hybrid or Huber)
      !    psiter_in : Number of Plane-search iterations on Alpha,Beta
      !
      real           :: rthr
      integer        :: psiter
      character(10)  :: norm
      optional       :: psiter,norm


      ps_norm="huber"
      if (present(norm)) then
         ps_norm=norm
      end if


      ps_iter=1
      if (present(psiter)) then
         ps_iter=psiter
      end if


      rt = abs(rthr)     ! Rt is positive
      if (ps_norm(1:5)=="huber" .or. ps_norm(1:6)=="hybrid") then
         if (rt==0.0)  call erexit('r_thr = 0. is not a valid choice.')
      endif

      ! write(0,*) " ===> Norm,  Rt, PS Iter = " ,ps_norm,rt,ps_iter

    end subroutine

    ! --------------------------------------------
    !              CGNORM function
    ! --------------------------------------------
    integer function cgnorm( forget, x, g, rr, gg)
      ! Iterative Plane-Search Stepper function 
      !    for different norms (Hybrid, Huber, etc.)

      ! Arguments: 
      ! x, g   : x, dx
      ! rr,gg  : R, dR
      ! forget : set steepest-decsend if true and CG-direction if false

      ! Output Flag (cgnorm): 
      !   2 : Exit
      !   1 : Exit with vanishing gradient status
      !   0 : Done without any exit or error
      !
      double precision, dimension(:), allocatable     :: sds, gdg, gds
      double precision, dimension(:), allocatable     :: c0,c1,c2
      real, dimension (:)  :: x, g, rr, gg
      logical              :: forget 
      double precision     :: alfa, beta, determ,c2dgg,c2dss,c2dgs,c1dg,c1ds
      integer              :: i,ps,stat

      cgnorm = 2
      allocate (c0(size(rr)),c1(size(rr)),c2(size(rr))) 
      allocate (gdg(size(rr)),sds(size(rr)),gds(size(rr))) 

      if( .not. allocated(s)) then
         forget = .true.
         allocate ( s(size( x))) 
         allocate (ss(size(rr))) 
      end if

      ! Loop over alpha , Beta
      do ps=1,ps_iter

         ! === Compute Norm and its derivatives
         if      (ps_norm(1:2)=="l2")     then
            stat = l2     (rr, c0,c1,c2)
         else if (ps_norm(1:2)=="l1")     then
            stat = l1     (rr, c0,c1,c2)
         else if (ps_norm(1:6)=="hybrid")  then
            stat = hybrid (rr, c0,c1,c2)
         else
            stat = huber  (rr, c0,c1,c2)
         end if

 
         ! === Pick Line Search or Plane Search
         if( forget) then
            ! Steepest Descent
            s    = 0.  
            ss   = 0.
            beta = 0.d0

            gdg   = dprod(gg,gg)
            c2dgg = sum(dprod(c2,gdg))
            
            !!! Discard C" if C"= 0  (OPTIONAL)
            ! if (sum(abs(c2))==0.0) c2dgg=sum(gdg)
               
            if( c2dgg == 0.0 )  then
               cgnorm = 1
               call erexit('NORM_CGSTEP: Gradient vanishes identically: C"*gg*gg=0 .')
            end if
            alfa = - sum( dprod(c1,gg)) / c2dgg

         else
            ! Plane Search for alpha & beta
            c1dg = - sum(dprod(c1,gg))   ! c1dg = -C'*G
            c1ds = - sum(dprod(c1,ss))   ! c1ds = -C'*S
            
            gdg = dprod(gg,gg)           ! search plane by solving 2-by-2
            sds = dprod(ss,ss)           ! C"*G . (G*alfa + S*beta) = -C'*G
            gds = dprod(gg,ss)           ! C"*S . (G*alfa + S*beta) = -C'*S

            c2dgg = sum(dprod(c2,gdg))   ! c2dgg = C"*G*G
            c2dss = sum(dprod(c2,sds))   ! c2dss = C"*S*S
            c2dgs = sum(dprod(c2,gds))   ! c2dgs = C"*G*S
             
            !if (sum(abs(c2))==0.0) then 
            ! !!! Discard C" if C"= 0 (OPTIONAL)
            ! c2dgg=sum(gdg)
            ! c2dss=sum(sds)
            !end if

            if ((c2dgg==0.d0) .OR. (c2dss==0.d0)) then
               cgnorm = 1;  
               write (0,*) 'Plane Search : Gradient vanishes identically: C"*gg*gg=0 OR C"*ss*ss=0.'
               return
            end if
            determ = c2dgg * c2dss * max (1.d0 - (c2dgs/c2dgg)*(c2dgs/c2dss), 1.d-12)
            alfa = ( c2dss * c1dg - c2dgs * c1ds ) / determ
            beta = (-c2dgs * c1dg + c2dgg * c1ds ) / determ
         end if

         !  Updates on model and residual
         s  = alfa *  g + beta *  s            ! update solution step
         ss = alfa * gg + beta * ss            ! update residual step
         x  =  x +  s                          ! update solution
         rr = rr + ss                          ! update residual
         forget = .false.;

         !write (0,*) "Alpha, Beta::" ,alfa,beta
         !write (0,*) "Determ::" ,determ
      end do


      deallocate (c0,c1,c2)
      deallocate (gdg,sds,gds) 
      cgnorm = 0
    end function

    ! --------------------------------------------
    !        De-Allocate 
    ! --------------------------------------------
    subroutine cgnorm_close( )
      if (allocated(s))  deallocate(s,ss)
    end subroutine
    
end module
